{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Array and Object (Dictionary)\n",
    "\n",
    "## Array\n",
    "\n",
    "### Declaration and Initialization\n",
    "\n",
    "- `Arrays` in JavaScript can be declared and initialized using square brackets `[]`.\n",
    "  ```js\n",
    "  let numbers = [1, 2, 3, 4, 5];\n",
    "  ```\n",
    "\n",
    "### Accessing Elements\n",
    "\n",
    "- Elements in an `array` can be accessed using square bracket notation with the index.\n",
    "  ```js\n",
    "  console.log(numbers[0]); // Output: 1\n",
    "  ```\n",
    "\n",
    "### Length Property\n",
    "\n",
    "- The `length` property returns the number of elements in an array.\n",
    "  ```js\n",
    "  console.log(numbers.length); // Output: 5\n",
    "  ```\n",
    "\n",
    "### Adding Elements\n",
    "\n",
    "- New elements can be added to the end of an array using the `push()` method.\n",
    "  ```js\n",
    "  numbers.push(6);\n",
    "  console.log(numbers); // Output: [1, 2, 3, 4, 5, 6]\n",
    "  ```\n",
    "\n",
    "### Removing Elements\n",
    "\n",
    "- Elements can be removed from the end of an array using the `pop()` method.\n",
    "  ```js\n",
    "  numbers.pop();\n",
    "  console.log(numbers); // Output: [1, 2, 3, 4, 5]\n",
    "  ```\n",
    "\n",
    "### Iterating Over Arrays\n",
    "\n",
    "- `Arrays` can be iterated over using loops like for loop or `forEach()` method.\n",
    "  ```js\n",
    "  for (let i = 0; i < numbers.length; i++) {\n",
    "    console.log(numbers[i]);\n",
    "  }\n",
    "  // Using forEach\n",
    "  numbers.forEach(function (number) {\n",
    "    console.log(number);\n",
    "  });\n",
    "  ```\n",
    "\n",
    "### Slicing Arrays\n",
    "\n",
    "- The `slice()` method returns a shallow copy of a portion of an `array` into a new `array` object.\n",
    "  ```js\n",
    "  let slicedArray = numbers.slice(2, 4);\n",
    "  console.log(slicedArray); // Output: [3, 4]\n",
    "  ```\n",
    "\n",
    "### Splicing Arrays:\n",
    "\n",
    "- The `splice()` method changes the contents of an `array` by removing or replacing existing elements and/or adding new elements in place.\n",
    "  ```js\n",
    "  console.log(numbers); // Output: [1, 2, 3, 4, 5]\n",
    "  numbers.splice(2, 1); // Removes 1 element from index 2\n",
    "  console.log(numbers); // Output: [1, 2, 4, 5]\n",
    "  ```\n",
    "\n",
    "### Concatenating Arrays:\n",
    "\n",
    "- The `concat()` method is used to merge two or more `arrays`.\n",
    "\n",
    "  ```js\n",
    "  let moreNumbers = [6, 7, 8];\n",
    "  let combinedArray = numbers.concat(moreNumbers);\n",
    "  console.log(combinedArray); // Output: [1, 2, 4, 5, 6, 7, 8]\n",
    "  ```\n",
    "\n",
    "### Finding Elements:\n",
    "\n",
    "- The `indexOf()` method returns the first index at which a given element can be found in the `array`, or `-1` if it is not present.\n",
    "  ```js\n",
    "  console.log(numbers.indexOf(4)); // Output: 2\n",
    "  ```\n",
    "\n",
    "### Checking Existence:\n",
    "\n",
    "- The `includes()` method determines whether an `array` includes a certain value among its entries, returning `true` or `false` as appropriate.\n",
    "\n",
    "  ```js\n",
    "  console.log(numbers.includes(3)); // Output: true\n",
    "  ```\n",
    "\n",
    "### Sorting Arrays:\n",
    "\n",
    "- The `sort()` method sorts the elements of an `array` in place and returns the sorted `array`.\n",
    "\n",
    "  ```js\n",
    "  numbers.sort();\n",
    "  console.log(numbers); // Output: [1, 2, 4, 5]\n",
    "  ```\n",
    "\n",
    "### Mapping Arrays:\n",
    "\n",
    "- The `map()` method creates a new `array` populated with the results of calling a provided function on every element in the calling `array`.\n",
    "\n",
    "  ```js\n",
    "  let doubledNumbers = numbers.map(function (number) {\n",
    "    return number * 2;\n",
    "  });\n",
    "  console.log(doubledNumbers); // Output: [2, 4, 8, 10]\n",
    "  ```\n",
    "\n",
    "### Filtering Arrays:\n",
    "\n",
    "- The `filter()` method creates a new `array` with all elements that pass the test implemented by the provided function.\n",
    "\n",
    "  ```js\n",
    "  let evenNumbers = numbers.filter(function (number) {\n",
    "    return number % 2 === 0;\n",
    "  });\n",
    "  console.log(evenNumbers); // Output: [2, 4]\n",
    "  ```\n",
    "\n",
    "### Reducing Arrays:\n",
    "\n",
    "- The `reduce()` method applies a function against an `accumulator` and each element in the `array` (from left to right) to reduce it to a single value.\n",
    "\n",
    "  ```js\n",
    "  let sum = numbers.reduce(function (accumulator, currentValue) {\n",
    "    return accumulator + currentValue;\n",
    "  }, 0);\n",
    "  console.log(sum); // Output: 12 (1 + 2 + 4 + 5)\n",
    "  ```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Object Literals (Dictionaries)\n",
    "\n",
    "### Declaration and Initialization\n",
    "\n",
    "- `Objects` in JavaScript are collections of `key-value pairs`, declared and initialized using curly braces `{}`.\n",
    "\n",
    "  ```js\n",
    "  let person = {\n",
    "    name: \"John\",\n",
    "    age: 30,\n",
    "    gender: \"male\",\n",
    "  };\n",
    "  ```\n",
    "\n",
    "### Accessing Properties\n",
    "\n",
    "- Properties of an `object` can be accessed using dot notation (`object.property`) or bracket notation (`object[\"property\"]`).\n",
    "\n",
    "  ```js\n",
    "  console.log(person.name); // Output: John\n",
    "  console.log(person[\"age\"]); // Output: 30\n",
    "  ```\n",
    "\n",
    "### Adding Properties\n",
    "\n",
    "- New properties can be added to an `object` simply by assigning a value to a new key.\n",
    "\n",
    "  ```js\n",
    "  person.email = \"john@example.com\";\n",
    "  console.log(person); // Output: { name: \"John\", age: 30, gender: \"male\", email: \"john@example.com\" }\n",
    "  ```\n",
    "\n",
    "### Removing Properties\n",
    "\n",
    "- Properties can be removed from an `object` using the `delete` keyword.\n",
    "\n",
    "  ```js\n",
    "  delete person.gender;\n",
    "  console.log(person); // Output: { name: \"John\", age: 30, email: \"john@example.com\" }\n",
    "  ```\n",
    "\n",
    "### Object Methods\n",
    "\n",
    "- `Objects` can contain `methods`, which are `functions` associated with the `object`.\n",
    "\n",
    "  ```js\n",
    "  let person = {\n",
    "    name: \"John\",\n",
    "    age: 30,\n",
    "    greet: function () {\n",
    "      console.log(\"Hello, my name is \" + this.name);\n",
    "    },\n",
    "  };\n",
    "  person.greet(); // Output: Hello, my name is John\n",
    "  ```\n",
    "\n",
    "### Object Iteration\n",
    "\n",
    "- `Objects` can be iterated over using `for...in` loop to access `keys` and `values`.\n",
    "\n",
    "  ```js\n",
    "  for (let key in person) {\n",
    "    console.log(key + \": \" + person[key]);\n",
    "  }\n",
    "  // Output:\n",
    "  // name: John\n",
    "  // age: 30\n",
    "  // email: john@example.com\n",
    "  ```\n",
    "\n",
    "### Nested Objects\n",
    "\n",
    "- `Objects` can contain other `objects` as `properties`, forming nested structures.\n",
    "\n",
    "  ```js\n",
    "  let person = {\n",
    "    name: \"John\",\n",
    "    address: {\n",
    "      city: \"New York\",\n",
    "      country: \"USA\",\n",
    "    },\n",
    "  };\n",
    "  console.log(person.address.city); // Output: New York\n",
    "  ```\n",
    "\n",
    "### Object Constructor\n",
    "\n",
    "- `Objects` can be created using constructor functions with the `new` keyword.\n",
    "\n",
    "  ```js\n",
    "  function Person(name, age) {\n",
    "    this.name = name;\n",
    "    this.age = age;\n",
    "  }\n",
    "  let john = new Person(\"John\", 30);\n",
    "  console.log(john.name); // Output: John\n",
    "  ```\n",
    "\n",
    "### Object Destructuring\n",
    "\n",
    "- Object destructuring allows you to extract multiple `properties` from an `object` and assign them to variables.\n",
    "\n",
    "```js\n",
    "let { name, age } = person;\n",
    "console.log(name); // Output: John\n",
    "console.log(age); // Output: 30\n",
    "```\n",
    "\n",
    "```js\n",
    "const person = {\n",
    "  firstName: \"John\",\n",
    "  lastName: \"Doe\",\n",
    "  age: 30,\n",
    "  hobbies: [\"music\", \"movies\", \"sports\"],\n",
    "  address: {\n",
    "    street: \"50 main st\",\n",
    "    city: \"Boston\",\n",
    "    state: \"MA\",\n",
    "  },\n",
    "};\n",
    "\n",
    "const {\n",
    "  firstName,\n",
    "  lastName,\n",
    "  address: { city },\n",
    "} = person;\n",
    "console.log(city); // Boston\n",
    "```\n",
    "\n",
    "### Transfer to JSON\n",
    "\n",
    "```js\n",
    "const person = {\n",
    "  firstName: \"John\",\n",
    "  lastName: \"Doe\",\n",
    "  age: 30,\n",
    "  hobbies: [\"music\", \"movies\", \"sports\"],\n",
    "  address: {\n",
    "    street: \"50 main st\",\n",
    "    city: \"Boston\",\n",
    "    state: \"MA\",\n",
    "  },\n",
    "};\n",
    "\n",
    "const personJSON = JSON.stringify(person);\n",
    "console.log(personJSON); //{\"firstName\":\"John\",\"lastName\":\"Doe\",\"age\":30,\"hobbies\":[\"music\",\"movies\",\"sports\"],\"address\":{\"street\":\"50 main st\",\"city\":\"Boston\",\"state\":\"MA\"}}\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## JavaScript Classes\n",
    "\n",
    "- JavaScript `classes` are a blueprint for creating objects with predefined `properties` and `methods`. They were introduced in **ES6 (ECMAScript 2015)** to provide a cleaner and more intuitive syntax for working with object-oriented programming compared to traditional constructor functions and prototypes.\n",
    "- Classes simplify the process of creating multiple objects with similar properties and behaviors. They use a combination of the `class` keyword, constructors, and methods to define and manipulate object instances.\n",
    "\n",
    "### Key Concepts\n",
    "\n",
    "- **Class Definition**: Use the `class` keyword followed by the class name. Inside the class, you define a `constructor` method to initialize new instances.\n",
    "- **Constructor Method**: A special method that is automatically called when a new instance of the class is created.\n",
    "- **Methods**: `Functions` defined inside a `class` to perform actions on the object. They can access the object's properties using this.\n",
    "- **Prototype Properties**: Properties defined on the prototype of a class that are shared across all instances.\n",
    "- **Overriding Properties**: `Instance properties` can override `prototype properties` by assigning new values directly to the instance.\n",
    "\n",
    "### Demo Code\n",
    "\n",
    "```js\n",
    "class Rabbit {\n",
    "constructor(type) {\n",
    "this.type = type;\n",
    "}\n",
    "\n",
    "speak(line) {\n",
    "console.log(`The ${this.type} rabbit says '${line}'`);\n",
    "}\n",
    "}\n",
    "let killerRabbit = new Rabbit(\"killer\");\n",
    "let blackRabbit = new Rabbit(\"black\");\n",
    "killerRabbit.speak('hi');\n",
    "// → The killer rabbit says 'hi'\n",
    "blackRabbit.speak('no');\n",
    "// → The black rabbit says 'no'\n",
    "\n",
    "Rabbit.prototype.teeth = \"small\";\n",
    "console.log(killerRabbit.teeth);\n",
    "// → small\n",
    "killerRabbit.teeth = \"long, sharp, and bloody\";\n",
    "console.log(killerRabbit.teeth);\n",
    "// → long, sharp, and bloody\n",
    "console.log(blackRabbit.teeth);\n",
    "// → small\n",
    "console.log(Rabbit.prototype.teeth);\n",
    "// → small\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Exercises\n",
    "\n",
    "### Flattening\n",
    "\n",
    "Use the `reduce` method in combination with the `concat` method to **“flatten” an array of arrays into a single array** that has all the elements of the original\n",
    "arrays.\n",
    "\n",
    "```js\n",
    "// Recursive function to flatten an array of arrays with multiple nesting levels\n",
    "function flattenArray(arr) {\n",
    "  return arr.reduce((acc, curr) => {\n",
    "    // Check if the current element is an array\n",
    "    if (Array.isArray(curr)) {\n",
    "      // Recursively flatten the nested array\n",
    "      return acc.concat(flattenArray(curr));\n",
    "    } else {\n",
    "      // If it's not an array, concatenate the element\n",
    "      return acc.concat(curr);\n",
    "    }\n",
    "  }, []);\n",
    "}\n",
    "\n",
    "// Example usage with nested arrays\n",
    "const arrayOfArrays = [[1, [2, [3,4], 45]], [3, 4], [5, 6]];\n",
    "console.log(flattenArray(arrayOfArrays)); //[1, 2, 3, 4, 45, 3, 4, 5, 6]\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Your own loop\n",
    "\n",
    "Write a **higher-order function `loop`** that provides something like a `for loop` statement. It takes a `value`, a `test` function, an `update` function, and a `body` function. \n",
    "Each iteration, it first runs the `test` function on the current loop value and stops if that returns false. Then it calls the `body` function, giving it the\n",
    "current value. Finally, it calls the `update` function to create a new value and starts from the beginning.\n",
    "\n",
    "\n",
    "```js\n",
    "function loop(value, test, update, body) {\n",
    "  // Use a while loop to perform the looping mechanism\n",
    "  while (test(value)) {\n",
    "    // Call the body function with the current value\n",
    "    body(value);\n",
    "    // Update the value for the next iteration\n",
    "    value = update(value);\n",
    "  }\n",
    "}\n",
    "\n",
    "// Example usage\n",
    "// This example starts with a value of 5, continues while the value is greater than 0,\n",
    "// prints the current value, and decrements the value by 1 each iteration.\n",
    "\n",
    "loop(\n",
    "  5,                      // Initial value\n",
    "  n => n > 0,             // Test function: continue if n > 0\n",
    "  n => n - 1,             // Update function: decrement n by 1\n",
    "  console.log             // Body function: print the current value\n",
    ");\n",
    "\n",
    "// Output:\n",
    "// 5\n",
    "// 4\n",
    "// 3\n",
    "// 2\n",
    "// 1\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "### Everything\n",
    "\n",
    "Analogous to the `some` method, arrays also have an `every` method.** This one returns `true` when the given function returns `true` for every element in the array.**\n",
    "In a way, `some` is a version of the `||` operator that acts on arrays, and `every` is like the `&&` operator.\n",
    "\n",
    "Implement `every` as a function that takes an `array` and a `predicate` function as parameters. Write two versions, one using a `loop` and one using the `some`\n",
    "method.\n",
    "\n",
    "```js\n",
    "// Implementation Using a Loop\n",
    "function everyLoop(array,predicate){\n",
    "    for(let element of array){\n",
    "        if(!predicate(element)){\n",
    "            return false;\n",
    "        }\n",
    "    }\n",
    "    return true;\n",
    "}\n",
    "\n",
    "// Example usage:\n",
    "console.log(everyLoop([1, 2, 3, 4], n => n > 0)); // Output: true\n",
    "console.log(everyLoop([1, 2, 3, -4], n => n > 0)); // Output: false\n",
    "\n",
    "\n",
    "// Implementation Using the some Method\n",
    "function everySome(array, predicate) {\n",
    "  return !array.some(element => !predicate(element));\n",
    "}\n",
    "\n",
    "// Example usage:\n",
    "console.log(everySome([1, 2, 3, 4], n => n > 0)); // Output: true\n",
    "console.log(everySome([1, 2, 3, -4], n => n > 0)); // Output: false\n",
    "```"
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
